"""
Gateway exceptions
"""
from typing import Optional

__all__ = (
    'TooManyRequests',
    'InvalidFunctionArguments',
    'CooldownException',
    'UsernameTooCommon',
    'ItemNotFound',
    'CheckFailed',
    'CommandNotFound',
    'MissingArgumentsException'
)


class TooManyRequests(Exception):
    """
    Calls when sending too many http requests
    """

    def __init__(self, message: Optional[str] = None) -> None:
        super().__init__(message or "You're sending too many requests")


class InvalidFunctionArguments(Exception):
    """
    Calls when you send invalid arguments to a function
    """

    def __init__(self, message: Optional[str] = None) -> None:
        super().__init__(message or "You sent an invalid argument to a function")


class CooldownException(Exception):
    """
    Calls when user execute a command whilst still on cooldown
    """

    def __init__(self, message: Optional[str] = None, *, user: str = '') -> None:
        super().__init__(message or f"User {user} is still on cooldown")


class UsernameTooCommon(Exception):
    """
    Calls when changing to a very common username
    """

    def __init__(self, message: Optional[str] = None) -> None:
        super().__init__(message or "Username is too common, try a different one")


class ItemNotFound(Exception):
    """
    Calls when searching for unknown item
    """

    def __init__(self, message: Optional[str] = None, *, item: Optional[str] = None) -> None:
        super().__init__(message or f"{item} not found")


class CheckFailed(Exception):
    """
    Calls when command check has failed
    """

    def __init__(self, message: Optional[str] = None, *, user: Optional[str] = None) -> None:
        super().__init__(message or f"Check failed for user {user or 'unknown'}")


class CommandNotFound(Exception):
    """
    Calls when command check has failed
    """

    def __init__(self, message: Optional[str] = None, *, command: Optional[str] = None) -> None:
        super().__init__(message or f"Command not found: {command}")


class MissingArgumentsException(Exception):
    """
    Calls when command missing required arguments
    """

    def __init__(self, message: Optional[str] = None, *, command: Optional[str] = None) -> None:
        super().__init__(message or f"Command {command} missing required arguments")



"""
The request handler.

Saves common useful functions for the api
"""
import time
from typing import Optional, Any, Union

from requests import Response, HTTPError

from discord_gateway.discord_objects import DiscordMessage

from discord_gateway.embed import Embed, DiscordFile
from discord_gateway.constants import Status, AllowedMentions
from discord_gateway.presence import DiscordActivity
from discord_gateway.exceptions import UsernameTooCommon

__all__ = (
    'RequestsManager',
)


class RequestsManager:
    from discord_gateway.api import HttpClient  # Circular imports

    def __init__(self, api: HttpClient):
        self.api = api

    async def send_message(self, channel_id: str, content: str = '', *, tts: bool = False, embed: Embed = None,
                           message_reference: Optional[str] = None,
                           allowed_mentions: Union[AllowedMentions, bool, None] = None,
                           file: Optional[DiscordFile] = None, threads: bool = False) -> Optional[DiscordMessage]:
        """
        Head sender for discord messages
        """
        payload: dict[str, Any] = {
            'content': content or None,
            'tts': tts,
            'embeds': [embed.to_json()] if embed else [],
            'message_reference':
                {
                    "channel_id": channel_id,
                    "message_id": message_reference,
                } if message_reference else {},
        }

        if allowed_mentions is False:
            payload['allowed_mentions'] = AllowedMentions.none().payload

        if allowed_mentions:
            payload['allowed_mentions'] = allowed_mentions.payload

        r = self.api.requests(method='POST', path=f'/channels/{channel_id}/messages',
                              json=payload, threads=threads, files=file.to_json() if file else None)
        if file:
            file.path.close()

        return DiscordMessage(r, client=self.api) if not threads else None

    @staticmethod
    async def get_presence(activity: Optional[DiscordActivity], status: Optional[Status], since: float = 0) -> dict:

        if status is Status.Offline:
            status = Status.Invisible

        elif status is Status.Idle:
            since = int(time.time() * 1000)

        activities = []

        if activity:
            activities = [activity.to_json()]

        payload = {
            'op': 3,
            'd': {
                'activities': activities,
                'afk': False,
                'since': since,
                'status': status.value,
            },
        }

        return payload

    async def edit_bot_user(
            self,
            payload: dict[str, Any]
    ) -> Union[Response, dict]:
        try:
            return self.api.requests(method='PATCH', path='/users/@me', json=payload)
        except HTTPError as e:
            if not e.args[0].split(':')[0] == '400 Client Error':
                raise e
            raise UsernameTooCommon()



import asyncio
import threading

from requests import Response
from typing import Optional, Iterator, Union, Any, AsyncIterator
from typing_extensions import Self

from discord_gateway.api import HttpClient
from discord_gateway.constants import ChannelType, Color, Permission, AllowedMentions
from discord_gateway.embed import Embed, DiscordFile
from discord_gateway.exceptions import InvalidFunctionArguments

__all__ = (
    'DiscordMessage',
    'DiscordChannel',
    'DiscordGuild',
)

from discord_gateway.guild_objects import DiscordRole


class DiscordMessage:
    __slots__ = '__token', 'data', 'id', 'content', 'api', 'guild', 'author', 'channel'

    def __init__(self, data: Union[dict, object], client: HttpClient) -> None:
        if not data:
            return
        from discord_gateway.discord_user import DiscordMember, DiscordUser  # Circular imports
        self.__token = client.get_token()
        data = dict(data.__dict__) if type(data) is not dict else data
        self.data = data
        self.id = data.get('id')
        self.content = data.get('content')
        self.api = client
        if data.get('guild_id', False):
            self.guild = DiscordGuild(guild_id=data.get('guild_id'), api=self.api)
            self.author = DiscordMember(data.get("member", {}), data.get("author", {}), self.guild.id,
                                        self.api)
        else:
            self.author: Union[DiscordMember, DiscordUser] = DiscordUser(data.get('author', {}), self.api)
        if data.get('channel_id', False):
            self.channel = DiscordChannel(channel_id=self.data.get('channel_id'), api=self.api)

    def __repr__(self) -> str:
        return self.id

    async def delete(self, threads: bool = False) -> Union[Response, dict]:
        """
        Deletes the message\n
        :param threads:
            Weather to use threading or not
        :return: :class:`Union[Response, dict]`
            The deleted message
        """
        return self.api.requests(method='DELETE', path=f'/channels/{self.channel.id}/messages/{self.id}',
                                 data={}, threads=threads)

    async def reply(self, content: str = '', *, tts: bool = False,
                    allowed_mentions: Union[AllowedMentions, bool] = None,
                    file: DiscordFile = None, embed: Embed = None) -> Self:
        """
        Quick way to reply to a message\n
        :param content: :class:`str`
            Message content
        :param tts: :class:`bool`
            Weather or not to use text to speech
        :param embed: :class:`Embed`
            Optional embed to send
        :param allowed_mentions:
            Weather the bot should mention everyone and roles via the command
        :param file:
            Additional attachment to the message
        :return: :class:`DiscordMessage`
            The replied message
        """
        return await self.api.manager.send_message(
            self.channel.id,
            content,
            tts=tts,
            embed=embed,
            allowed_mentions=allowed_mentions,
            file=file,
            message_reference=self.id,
        )

    async def edit(self, content: str) -> Union[Response, dict]:
        """
        Edit message\n
        :param content: :class:`str`
            New content
        :return: :class:`Union[Response, dict]`
            New message
        """
        headers = {
            'Authorization': self.api.get_auth(),
            'Content-Type': 'application/json'
        }
        return self.api.requests(method='PATCH', path=f'/channels/{self.channel.id}/messages/{self.id}',
                                 headers=headers, data='{"content":"' + content + '"}')


class DiscordChannel:
    __slots__ = 'id', '__token', 'api', '_data'

    def __init__(self, channel_id: str, api: HttpClient, use_data: bool = False) -> None:
        self.id: str = channel_id
        self.__token = api.get_token()
        self.api = api
        self._data: dict = {}
        if use_data:
            threading.Thread(target=asyncio.run, args=(self.get_data(),)).start()

    def __repr__(self) -> str:
        return self.id

    async def get_data(self) -> None:
        self._data = self.api.requests(method='GET', path=f'/channels/{self.id}')

    async def edit(self, name: Optional[str] = None, type: ChannelType = None, position: Optional[int] = None,
                   topic: Optional[str] = None) -> None:
        """
        Modify a channel
        :param name:
        :param type:
        :param position:
        :param topic:
        :return:
        """
        await self.get_data()
        payload = {
            'name': name or self.name,
            'type': type or self._data.get('type'),
            'position': position or self._data.get('position'),
            'topic': topic or self._data.get('topic'),
        }

        self.api.requests('PATCH', f'/channels/{self.id}', json=payload)

    async def send(self, content: str = '', *, tts: bool = False, message_reference: str = '', embed: Embed = None,
                   threads: bool = False, allowed_mentions: Union[AllowedMentions, bool] = None,
                   file: DiscordFile = None) -> DiscordMessage:
        """
        Send message in channel\n
        :param content: :class:`str`
            The content to send
        :param tts: :class:`bool`
            Weather or not to use text to speach on send
        :param message_reference:
            The message to reply to
        :type message_reference: str
        :param embed: :class:`Embed`
            An embed to send with the message
        :param allowed_mentions:
            Weather the bot should mention everyone and roles via the command
        :param file:
            Additional attachment to the message
        :param threads: :class:`bool`
            Weather to use threads
        :return: :class:`Optional[DiscordMessage]`
            The final message (only works without threading)
        """
        return await self.api.manager.send_message(
            self.id,
            content,
            tts=tts,
            embed=embed,
            message_reference=message_reference,
            allowed_mentions=allowed_mentions,
            file=file,
            threads=threads
        )

    async def delete(self, threads: bool = False) -> None:
        """
        Delete the channel\n
        :param threads:
            Weather to use threading or not.
        """
        self.api.requests(method='DELETE', path=f'/channels/{self.id}', threads=threads)

    async def generate_invite(self, *, max_age: int = 0, max_uses: int = 0,
                              temporary: bool = False) -> Union[str, dict]:
        """
        Creates an invitation to the server\n
        :param max_age: :class:`int`
            Max age of the invite before it gets deleted
        :param max_uses: :class:`int`
            Max allowed uses of the invite
        :param temporary: :class:`bool`
            Grants temporary membership on join
        :return: :class:`Union[str, dict]`
            The generated invite
        """
        payload = {
            "max_age": max_age,
            "max_uses": max_uses,
            "temporary": temporary,
        }
        return self.api.requests(method='POST', path=f'/channels/{self.id}/invites', json=payload)

    async def purge(self, limit: int = 50) -> None:
        """
        Bulk delete messages in a channel
        :param limit:
        :return:
        """
        messages = self.fetch_messages(limit=limit)

        payload = {
            'messages': [message.id async for message in messages]
        }

        self.api.requests(method='POST', path=f'/channels/{self.id}/messages/bulk-delete', json=payload)

    async def fetch_messages(self, limit: int = 50) -> AsyncIterator[DiscordMessage]:
        """

        :param limit:
        :return:
        """
        messages = self.api.requests(method='GET', path=f'/channels/{self.id}/messages?limit={limit}')
        for msg in messages:
            yield DiscordMessage(msg, client=self.api)

    @property
    def messages(self) -> Iterator[DiscordMessage]:
        """
        Get channel messages\n
        :return: :class:`DiscordMessage`
            Every message within 100 messages limit (use async fetch messages for more or less than the limit)
        """
        r = self.api.requests(method='GET', path=f'/channels/{self.id}/messages?limit=100')
        for msg in r:
            yield DiscordMessage(msg, client=self.api)

    @property
    def name(self) -> str:
        """

        :return:
        """
        return self.api.requests(method='GET', path=f'/channels/{self.id}').get('name')


class DiscordGuild:
    __slots__ = 'id', '__token', 'api'

    def __init__(self, guild_id: str, api: HttpClient) -> None:
        self.id: str = guild_id
        self.__token = api.get_token()
        self.api = api

    def __repr__(self) -> str:
        return self.id

    @property
    def name(self) -> str:
        return self.api.requests(method='GET', path=f'/guilds/{self.id}').get('name')

    @property
    def channels(self) -> Iterator[DiscordChannel]:
        """
        Get all guild channels\n
        :return: :class:`DiscordChannel`
            The channels of the guild
        """
        for channel in self.api.requests(method='GET', path=f'/guilds/{self.id}/channels'):
            yield DiscordChannel(channel.get('id'), api=self.api, use_data=False)

    @property
    def members(self) -> object:
        from discord_gateway.discord_user import DiscordMember  # Circular import

        for member in self.api.requests(method='GET', path=f'/guilds/{self.id}/members?limit=1000'):
            yield DiscordMember(member, member.get('user'), self.id, client=self.api)

    @property
    def roles(self) -> Iterator[DiscordRole]:
        """
        Get all guild roles
        :return:
            List of role ids
        """
        for role in self.api.requests(method='GET', path=f'/guilds/{self.id}/roles'):
            yield DiscordRole(role, self.id, client=self.api)

    @property
    def data(self) -> dict:
        return self.api.requests(method='GET', path=f'/guilds/{self.id}')

    async def edit(
            self, name: Optional[str] = None,
            *,
            region: Optional[str] = None,
            icon: Any = None,
            banner: Any = None,
            description: Optional[str] = None,
            **kwargs
    ) -> Union[Response, dict]:
        """
        Edit a guild
        :param name:
        :param region:
        :param icon:
        :param banner:
        :param description:
        :param kwargs:
        :return:
        """
        data = self.data

        payload = {
            'name': name or self.name,
            'region': region or data.get('region'),
            'icon': icon or data.get('icon'),
            'banner': banner or data.get('banner'),
            'description': description or data.get('description'),
        }

        if kwargs:
            payload += {
                key: value or data.get(key) for key, value in kwargs
            }

        return self.api.requests(method='PATCH', path=f'/guilds/{self.id}', json=payload)

    def get_role(self, role_id: Optional[str] = None, *, name: Optional[str] = None,
                 role_list: Optional[list[DiscordRole]] = None) -> Optional[DiscordRole]:
        """
        Get a role by id or name
        :param role_id:
        :param name:
        :param role_list:
        :return:
        """
        roles = role_list or self.roles  # To save spam requests
        try:
            if role_id is not None:
                return next(role for role in roles if role_id == role.id)
            if name is None:
                raise InvalidFunctionArguments('You must send 1 argument for the search query')
            return next(role for role in roles if name == role.name.lower())
        except StopIteration:
            return None

    async def create_role(
            self,
            *,
            name: str = None,
            permissions: Optional[Permission] = '0',
            color: Optional[Color] = None,
            hoist: Optional[bool] = None,
            mentionable: Optional[bool] = None,
    ) -> DiscordRole:
        """
        Creates a new role
        :param name:
        :param permissions:
        :param color:
        :param hoist:
        :param mentionable:
        :return:
        """

        color = color or Color.Default

        if type(permissions) is not int:
            permissions = permissions.value

        payload: dict[str, Any] = {
            'name': name or 'new role',
            'permissions': str(permissions),
            'color': color.value,
            'mentionable': mentionable or False
        }

        if hoist is not None:
            payload['hoist'] = hoist

        data = self.api.requests('POST', f'/guilds/{self.id}/roles', json=payload)
        return DiscordRole(data, self.id, client=self.api)

    async def create_channel(self, name: str, *, channel_type: ChannelType = ChannelType.TextChannel,
                             permission_overwrites: list = (), threads: bool = False) -> Optional[DiscordChannel]:
        """
        Create a channel in the guild\n
        :param name: :class:`str`
            The name of the channel
        :param channel_type: :class:`ChannelType`
            The channel type
        :param permission_overwrites: :class:`list`
            The permissions of the channel
        :param threads: :class:`bool`
            Weather to use threads or normal async
        :return: :class:`DiscordChannel`
            The created channel
        """
        payload = {
            'name': '-'.join(name.split(' ')),
            'permission_overwrites': permission_overwrites,
            'type': channel_type.value
        }
        data = self.api.requests(method='POST', path=f'/guilds/{self.id}/channels', json=payload, threads=threads)
        if not threads:
            return DiscordChannel(data.get('id'), api=self.api)

    async def leave(self) -> None:
        """
        Leaves the guild\n
        """
        self.api.requests(method='DELETE', path=f'users/@me/guilds/{self.id}')

    from discord_gateway.discord_user import DiscordMember  # Circular import

    async def get_member(self, member_id: str) -> Optional[DiscordMember]:
        """
        Get discord member by id\n
        :param member_id:
            The id of the member
        :type member_id: str
        :return:
            The gotten member
        :rtype: DiscordMember
        """

        from discord_gateway.discord_user import DiscordMember  # Circular import + Not a type hint

        data = self.api.requests(method='GET', path=f'/guilds/{self.id}/members/{member_id}')
        if data:
            return DiscordMember(data, data.get('user'), self.id, client=self.api)

    async def kick(self, member: DiscordMember) -> None:
        """
        Kick a member from the guild\n
        :param member: :class:`DiscordMember`
            The member to kick from the guild
        """
        self.api.requests(method='DELETE', path=f'/guilds/{self.id}/members/{member.id}')

    async def ban(self, member: DiscordMember, delete_message_seconds: int = 3600) -> None:
        """
        Ban a member from the guild\n
        :param member: :class:`DiscordMember`
            The member to ban from the guild
        :param delete_message_seconds: :class:`int`
            Delete message seconds, delete all the messages from the user
            in the given time frame
        """
        payload = {
            "delete_message_seconds": delete_message_seconds
        }
        self.api.requests(method='PUT', path=f'/guilds/{self.id}/bans/{member.id}',
                          data=payload)

    async def unban(self, member_id: str) -> None:
        """
        Unbans a member from the guild\n
        :param member_id: :class:`str`
            The member id to unban
        """
        self.api.requests(method='DELETE', path=f'/guilds/{self.id}/bans/{member_id}')



from types import SimpleNamespace
from typing import Union, Iterator

from requests import Response

from discord_gateway.api import HttpClient
from discord_gateway.discord_objects import DiscordMessage
from discord_gateway.guild_objects import DiscordRole
from discord_gateway.constants import Permission

__all__ = (
    'UserMessage',
    'DiscordUser',
    'DiscordMember'
)


class UserMessage(DiscordMessage):

    def __init__(self, data: object, client: HttpClient) -> None:
        try:  # Try obtaining data
            super().__init__(data, client)
        except AttributeError:
            pass
        self.author = DiscordUser(self.author, self.__token)


class DiscordUser:
    __slots__ = 'id', 'name', 'discriminator', 'mention', 'bot', '__token', 'api'

    def __init__(self, user: dict, client: HttpClient) -> None:
        self.id: str = user.get('id')
        self.name: str = user.get('username')
        self.discriminator: str = user.get('discriminator')
        self.mention: str = '<@%s>' % self.id
        self.bot: bool = user.get('bot', False)
        self.__token: str = client.get_token()
        self.api = client

    def __repr__(self) -> str:
        return f'{self.name}#{self.discriminator}'

    async def send(self, *content) -> UserMessage:
        """
        Send dm to user\n
        :param content: :class:`list[str]`
            The content to send
        :return: :class:`UserMessage`
            The message
        """
        headers = {
            'Authorization': self.api.get_auth(),
            'Content-Type': 'application/json'
        }
        user_id = self.api.requests(method='POST', path=f'/users/@me/channels', json={"recipients": [self.id]},
                                    headers=headers).get('id')
        if content:
            r = self.api.requests(method='POST', path=f'/channels/{user_id}/messages', data={'content': content})
            return UserMessage(SimpleNamespace(**r), client=self.api)


class DiscordMember(DiscordUser):
    __slots__ = 'guild_id', 'nick', 'token', '_data'

    def __init__(self, member: dict, user: dict, guild_id: str, client: HttpClient) -> None:
        super().__init__(user, client)
        # Combining data
        self._data = member
        self._data.update(user)
        # Setting variables
        self.guild_id: str = guild_id
        self.nick: str = member.get("nick", "")
        self.token: str = client.get_token()

    @property
    def roles(self) -> Iterator[DiscordRole]:
        from discord_gateway.discord_objects import DiscordGuild

        guild = DiscordGuild(self.guild_id, api=self.api)
        roles = list(guild.roles)
        return (guild.get_role(r, role_list=roles) for r in self._data.get('roles'))

    def add_role(self, role_id: str) -> Union[Response, dict]:
        """
        Add a role to member\n
        :param role_id: :class:`str`
            The role id
        :return: :class:`Union[Response, dict]`
            The added role
        """
        return self.api.requests(method='PUT', path=f'/guilds/{self.guild_id}/members/{self.id}/roles/{role_id}')

    def remove_role(self, role_id: str) -> None:
        """
        Remove a role from member\n
        :param role_id: :class:`str`
            The role id
        """
        self.api.requests(method='DELETE', path=f'/guilds/{self.guild_id}/members/{self.id}/roles/{role_id}')

    @property
    def permissions(self) -> Iterator[Permission]:
        from discord_gateway.discord_objects import DiscordGuild

        data = DiscordGuild(self.guild_id, api=self.api).data

        if data.get('owner_id') == self.id:
            return (perm for perm in list(Permission))

        base = 0

        for role in self.roles:
            base |= sum(p.value for p in role.permissions)

        return (perm for perm in Permission.from_int(base))

    def has_permission(self, permission: Permission) -> bool:
        return permission in list(self.permissions)

    async def kick(self) -> None:
        """
        Kicks the member\n
        """
        self.api.requests(method='DELETE', path=f'/guilds/{self.guild_id}/members/{self.id}')

    async def ban(self, delete_message_seconds: int = 3600) -> None:
        """
        Bans the member\n
        :param delete_message_seconds: :class:`int`
            Delete message seconds, delete all the messages from the user
            in the given time frame
        """
        payload = {
            "delete_message_seconds": delete_message_seconds
        }
        self.api.requests(method='PUT', path=f'/guilds/{self.guild_id}/bans/{self.id}', data=payload)




"""
Discord Presence classes
"""

from typing import Union

from discord_gateway import ActivityType


__all__ = (
    'DiscordActivity',
)


class DiscordActivity:

    """
    Class for discord activities displayed in profile and in status
    """

    __slots__ = 'type', 'name', 'url', 'assets'

    def __init__(self, type: Union[ActivityType, int], name: str = '', url: str = '', assets: dict = ()) -> None:
        self.type = type.value
        self.name = name or None
        self.url = url or None
        self.assets = assets

    def to_json(self) -> dict:
        """
        Converts the activity data to dict\n
        :return:
            Activity data as a dict
        """
        return {
            key: getattr(self, key)
            for key in self.__slots__
            if hasattr(self, key)
        }



"""
Command object
"""

__all__ = (
    'ClientCommand',
    'CommandThread',
)

import asyncio
import threading
import time
from typing import Callable, Optional, Any
from discord_gateway import CooldownException
from .command_context import CommandContext


class CommandThread(threading.Thread):

    def __init__(self, target, args=None, kwargs=None) -> None:
        threading.Thread.__init__(self, target=target, args=args, kwargs=kwargs)
        self.exc: Optional[BaseException] = None

    def run(self) -> None:
        self.exc = None
        try:
            self._target(*self._args, **self._kwargs)
        except BaseException as e:
            self.exc = e

    def join(self, timeout=None) -> None:
        super(CommandThread, self).join(timeout)
        if self.exc:
            raise self.exc


class ClientCommand:
    __slots__ = 'function', 'description', 'cooldown', '_check', 'track_cooldown', 'error_allocator'

    def __init__(self, function: Callable, description: str, cooldown: int = 0) -> None:
        self.function = function
        self.description = description
        self.cooldown = cooldown
        self._check: Optional[Callable] = None
        self.track_cooldown: dict[str, float] = {}
        self.error_allocator: Optional[Callable] = None

    def error(self, allocator: Callable) -> Callable[..., Any]:
        self.error_allocator = allocator
        return allocator

    def dispatch(self, context: CommandContext, args: list, bot: Optional[Any] = None) -> None:
        user_id: str = context.author.id
        if self.cooldown:
            if user_id not in self.track_cooldown:
                self.track_cooldown[user_id] = time.time()
            elif time.time() - self.track_cooldown[user_id] < self.cooldown:
                context.cooldown = self.get_cooldown
                if not self.error_allocator:
                    return bot.raise_error(context, CooldownException(user=user_id))
                return threading.Thread(target=asyncio.run,
                                        args=(self.error_allocator(context, CooldownException(user=user_id)),)).start()
            else:
                self.track_cooldown[user_id] = time.time()
        command_thread = CommandThread(target=asyncio.run, args=(self.function(context, *args),))
        # Running thread
        command_thread.start()
        try:
            command_thread.join()
        except BaseException as e:
            if not self.error_allocator:
                return bot.raise_error(context, e)
            threading.Thread(target=asyncio.run, args=(self.error_allocator(context, e),)).start()

    def get_cooldown(self, user_id: str) -> Optional[float]:
        if user_id not in self.track_cooldown or not self.cooldown:
            return None
        return self.cooldown - (time.time() - self.track_cooldown.get(user_id))

    def set_check(self, check: Callable) -> None:
        self._check = check

    def eval_check(self, ctx: CommandContext) -> bool:
        if self._check is None:
            return True
        return self._check(ctx)



from typing import Optional, Union, Callable

from discord_gateway import Embed, DiscordFile
from discord_gateway.api import HttpClient
from discord_gateway.discord_objects import DiscordMessage

__all__ = (
    'CommandContext',
)


class CommandContext(DiscordMessage):

    def __init__(self, data: Union[dict, object], client: HttpClient) -> None:
        super().__init__(data, client)
        self.api = client
        self._cooldown: Optional[Callable] = None

    @property
    def cooldown(self) -> float:
        if not self._cooldown:
            return 0
        return self._cooldown(self.author.id)

    @cooldown.setter
    def cooldown(self, value: Callable) -> None:
        self._cooldown = value

    from discord_gateway.constants import AllowedMentions

    async def send(
        self,
        content: str = '',
        *,
        tts: bool = False,
        embed: Embed = None,
        message_reference: Optional[str] = None,
        allowed_mentions: Union[AllowedMentions, bool] = None,
        file: DiscordFile = None,
        threads: bool = False
    ) -> Optional[DiscordMessage]:
        """
        Faster way to respond to a command\n
        :param content: :class:`str`
            The content to send
        :param tts: :class:`bool`
            Weather or not to use text to speach on send
        :param message_reference:
            The message to reply to
        :type message_reference: str
        :param embed: :class:`Embed`
            An embed to send with the message
        :param allowed_mentions:
            Weather the bot should mention everyone and roles via the command
        :param file:
            Additional attachment to the message
        :param threads: :class:`bool`
            Weather to use threads
        :return: :class:`Optional[DiscordMessage]`
            The final message (only works without threading)
        """
        return await self.api.manager.send_message(
            self.channel.id,
            content,
            tts=tts,
            embed=embed,
            message_reference=message_reference,
            allowed_mentions=allowed_mentions,
            file=file,
            threads=threads
        )



import logging as log
import json
import threading
from typing import Optional, Union

import requests
from requests import Response, JSONDecodeError

DISCORD_API = 'https://discord.com/api/v9'  # Current gateway api

__all__ = (
    'HttpClient',
)


def handle_api_response(resp: requests.Response) -> Union[Response, dict]:
    resp.raise_for_status()  # Raise errors
    try:
        body: json = resp.json()
        if 'errors' in body:
            raise Exception(body)
        return body
    except TypeError:
        return resp
    except JSONDecodeError:
        return resp


class HttpClient:
    def __init__(self, token: str, bot: bool = True) -> None:

        from discord_gateway.request_manager import RequestsManager  # Circular imports

        self._token: str = token
        self.is_bot = bot
        self.headers = {'Authorization': f'{"Bot " if bot else ""}{self._token}'}
        self.manager = RequestsManager(self)  # Request manager

    def requests(self, method: str, path: str, *, data: Union[str, dict, None] = None, headers: Optional[dict] = None,
                 threads: bool = False,
                 **kwargs) -> Union[Response, dict, None]:
        """
        Send a request using self api\n
        :param method: :class:`str`
            The method to use {uppercase}
        :param path: :class:`str`
            The path of the request
        :param data: :class:`Optional[Union[dict, str]]`
            Additional data for the requests
        :param headers: :class:`Optional[dict]`
            Headers to override
        :param threads: :class:`bool`
            Weather to use threads when requesting
        :param kwargs: :class:`dict`
            Additional params
        :return: :class:`Union[Response, dict]
            The response
        """
        url = f'{DISCORD_API}{path}'

        headers = headers or self.headers

        if threads:
            return threading.Thread(target=self.send, args=(method, url, data, headers), kwargs={**kwargs}).start()
        else:
            return self.send(method, url, data, headers, **kwargs)

    @staticmethod
    def send(method: str, url: str, data: Union[str, dict, None], headers: Optional[dict],
             **kwargs) -> Union[Response, dict, bool]:
        if 'json' in kwargs:
            log.debug(f'Sending payload - {kwargs["json"]} via method - {method}')
        elif data:
            log.debug(f'Sending payload - {data} via method - {method}')
        try:
            if method == "GET":
                resp = requests.get(url, headers=headers, **kwargs)
                return handle_api_response(resp)
            elif method == "PUT":
                resp = requests.put(url, headers=headers, **kwargs)
                return handle_api_response(resp)
            elif method == "POST":
                resp = requests.post(url, headers=headers, data=data, **kwargs)
                return handle_api_response(resp)
            elif method == "DELETE":
                resp = requests.delete(url, headers=headers, data=data, **kwargs)
                return handle_api_response(resp)
            elif method == "PATCH":
                session = requests.Session()
                resp = session.patch(url, headers=headers, data=data, **kwargs)
                return handle_api_response(resp)

            else:
                raise Exception("Unsupported HTTP method {method}")

        except requests.exceptions.HTTPError as e:

            from discord_gateway.exceptions import TooManyRequests

            match e.args[0].split(':')[0]:
                case '429 Client Error':
                    raise TooManyRequests()
                case _:
                    raise e

    def get_token(self) -> str:
        """
        Gets the api token\n
        :return: :class:`str`
            API token
        """
        return self._token

    def get_auth(self) -> str:
        """
        Get auth header\n
        :return: :class:`str`
            The formatted authorize header
        """
        return f'{"Bot " if self.is_bot else ""}{self._token}'



from typing import Optional, Callable, Any, Union, AsyncIterator, Iterator

from requests import Response

from discord_gateway.discord_objects import DiscordChannel, DiscordGuild
from discord_gateway.discord_user import DiscordUser, DiscordMember
from discord_gateway.gateway_handler import Gateway
from discord_gateway.gateway import GatewayMessage
from discord_gateway.api import HttpClient
from types import SimpleNamespace
from discord_gateway.constants import Intents, Status
from discord_gateway.presence import DiscordActivity

__all__ = (
    'GatewayBot',
    'ClientUser'
)


class GatewayBot(Gateway):

    def __init__(self, *, intents: int = Intents.All, allow_dms: bool = True) -> None:
        super().__init__(intents=intents.value, allow_dms=allow_dms)
        self.api: Optional[HttpClient] = None
        self.user: Optional[ClientUser] = None
        self.add_listener(self.default_ready, event='ready', type='default')

    async def default_ready(self, msg: GatewayMessage) -> None:
        """
        Default ready,\n
        Stores user data, api and user tag.\n
        :param msg: :class:`GatewayMessage`
            The data received from the identify response
        """

        self.api = HttpClient(self._token, bot=self.is_bot)
        self.user = ClientUser(msg, self.allow_dms, client=self.api)

    def event(self, func: Callable) -> Callable[..., Any]:
        """
        **Decorator**\n
        Used to assign a function with an event\n
        :param func: The function to assign
        :return: :class:`Callable`
            Returns the function itself
        """
        self.add_listener(func, event=func.__name__.replace('on_', '', 1) if func.__name__.startswith(
            'on_') else func.__name__)
        return func

    @property
    def dms(self) -> Union[Response, dict]:
        """
        Get open dms of bot\n
        :return: :class:`Union[Response, dict]`
            Returns the open dms
        """
        return self.api.requests(method='GET', path='/users/@me/channels')

    @property
    def guilds(self) -> Iterator[DiscordGuild]:
        """
        Get current user guilds\n
        :return:
            Iterator of the guilds
        """
        for guild in self.api.requests(method='GET', path='/users/@me/guilds'):
            yield DiscordGuild(guild.get('id'), api=self.api)

    async def get_channel(self, channel_id: str) -> DiscordChannel:
        """
        Get channel by his id\n
        :param channel_id: :class:`str`
            The channel id
        :return: :class:`DiscordChannel`
            The channel
        """
        return DiscordChannel(channel_id, api=self.api)

    async def get_guild(self, guild_id: str) -> DiscordGuild:
        """
        Get guild by id\n
        :param guild_id: :class:`str`
            The guild id
        :return: :class:`DiscordGuild`
            Returns the guild
        """
        return DiscordGuild(guild_id, api=self.api)

    async def get_user(self, user_id: str) -> DiscordUser:
        """
        Get user by his id\n
        :param user_id: :class:`str`
            The user id
        :return: :class:`DiscordUser`
            Returns the user
        """
        headers = {
            'authorization': self.api.get_auth(),
            'content-type': 'application/json'
        }
        return DiscordUser(self.api.requests(method='POST', path=f'/users/@me/channels', json={"recipients": [user_id]},
                                             headers=headers), self.api)

    async def change_presence(self, activity: Optional[DiscordActivity] = None,
                              status: Union[Status, str] = Status.Online) -> None:
        """
        Change self presence\n
        :param activity:
            The activity to show
        :type activity: Optional[DiscordActivity]
        :param status:
            The status of bot ex. idle
        :type status: Union[Status, str]
        """
        payload = await self.api.manager.get_presence(activity, status)
        await self.send(payload)

    async def join(self, server_invite: str) -> Union[Response, dict]:
        """
        Join a guild by invite {SelfBot only}\n
        :param server_invite: :class:`str`
            The server invite link
        :return: :class:`Union[Response, dict]`
            The obtained server
        """
        return self.api.requests(method='POST', path=f'/invites/{server_invite}')

    async def leave(self, guild_id: str) -> None:
        """
        Leaves a guild\n
        :param guild_id: :class:`str`
            The guild id to leave
        """
        self.api.requests(method='DELETE', path=f'/users/@me/guilds/{guild_id}')

    async def create_group_chat(self, ids: list[str]) -> Union[Response, dict]:
        """
        Creates a group chat with members {SelfBot only}\n
        :param ids: :class:`list[str]`
            The list of the member ids
        :return: :class:`Union[Response, dict]`
            The opened group chat
        """
        headers = {
            'authorization': self.api.get_auth(),
            'content-type': 'application/json'
        }
        return self.api.requests(method='POST', path=f'/users/@me/channels', json={'recipients': ids}, headers=headers)

    async def members_from_guilds(self, guild_id: str) -> AsyncIterator[DiscordMember]:
        """
        Get possible members from a guild
        :param guild_id:
        :return:
        """
        if guild_id not in [guild.get('id') for guild in self.user.guilds]:
            raise AttributeError('Guild not found')
        for member in list(filter(lambda guild: guild.get('id') == guild_id, self.user.guilds))[0]['members']:
            yield DiscordMember(member, member.get('user'), guild_id, client=self.api)


class ClientUser(SimpleNamespace):
    __slots__ = 'api', 'id', 'username', 'discriminator', 'bot', 'allow_dms', 'guilds', 'tag'

    def __init__(self, msg: GatewayMessage, allow_dms: bool, client: HttpClient):
        super(ClientUser, self).__init__(**msg.data.user)
        self.api = client
        self.id: str = msg.data.user.get('id')
        self.username: str = msg.data.user.get('username')
        self.discriminator: str = msg.data.user.get('discriminator')
        self.avatar: Any = msg.data.user.get('avatar')
        self.bot: bool = msg.data.user.get('bot', False)
        self.allow_dms = allow_dms  # Check if dms are allowed in config
        self.tag = f'{self.username}#{self.discriminator}'

    async def edit(
            self,
            username: Optional[str] = None,
            *,
            avatar: Any = None
    ) -> Union[Response, dict]:
        """
        Edit a user profile\n
        :param username:
            New username
        :type username: Optional[str]
        :param avatar:
            New pfp (profile picture/avatar)
        :type avatar: Any
        :return
            The new updated user
        """
        payload: dict[str, Any] = {'username': (username or self.username)}

        if avatar:
            payload[avatar] = avatar

        return await self.api.manager.edit_bot_user(payload=payload)



from types import SimpleNamespace
from typing import Optional

from typing_extensions import Self

from discord_gateway.api import HttpClient
from discord_gateway.constants import Permission

__all__ = (
    'DiscordRole',
)


class DiscordRole:
    __slots__ = 'id', 'name', 'color', 'api', 'data', '__token', 'guild_id', 'permissions', 'position'

    def __init__(self, data: dict, guild_id: str, client: HttpClient) -> None:
        self.id: str = data.get('id')
        self.name: str = data.get('name')
        self.permissions: list[Permission] = list(Permission.from_int(data.get('permissions', 0)))
        if Permission.Administrator in self.permissions:
            self.permissions = list(Permission)
        self.color: int = data.get('color', 0)
        self.position: int = data.get('position', 0)
        self.data = SimpleNamespace(**data)
        self.guild_id = guild_id
        self.api = client
        self.__token = client.get_token()

    def __repr__(self) -> str:
        return self.name

    async def delete(self) -> None:
        self.api.requests(method='DELETE', path=f'/guilds/{self.guild_id}/roles/{self.id}')

    async def edit(
            self,
            name: Optional[str] = None,
            *,
            permissions: Optional[Permission] = None,
            color: Optional[int] = None,
            hoist: Optional[bool] = None,
            unicode_emoji: Optional[str] = None,
            mentionable: Optional[bool] = None
    ) -> Self:
        """

        :param name:
        :param permissions:
        :param color:
        :param hoist:
        :param unicode_emoji:
        :param mentionable:
        :return:
        """

        if type(permissions) is not int:
            permissions = permissions.value

        payload = {
            'name': name or self.name,
            'permissions': str(permissions) or self.permissions,
            'color': color or self.color,
            'hoist': hoist or self.data.hoist,
            'unicode_emoji': unicode_emoji or self.data.unicode_emoji,
            'mentionable': mentionable or self.data.mentionable,
        }

        data = self.api.requests(method='PATCH', path=f'/guilds/{self.guild_id}/roles/{self.id}', json=payload)

        return self.__init__(data, self.guild_id, client=self.api)



from __future__ import annotations

import asyncio
import inspect
import re
import threading
from inspect import signature
from typing import Callable, Optional, Any

import discord_gateway
from .command_context import CommandContext
from .command import ClientCommand
from discord_gateway import GatewayBot, GatewayMessage, Embed, Color, DiscordUser, DiscordMember, CommandNotFound
from discord_gateway.exceptions import ItemNotFound, CheckFailed

from itertools import zip_longest
from requests import HTTPError
import logging as log

__all__ = (
    'CommandsBot',
)


async def _conv(bot: GatewayBot, ctx: CommandContext, item: str, anno: type) -> Any:
    """
    Converts annotations to items\n
    :param bot:
        The user object
    :param ctx:
        The message context
    :param item:
        The item to convert
    :param anno:
        The annotation
    :return:
        Item after convertion
    """
    if anno is DiscordUser:
        try:
            if type(item) is inspect.Parameter:
                return item.default
            return await bot.get_user(re.sub(r'[@!<>]', '', item))
        except HTTPError:
            raise ItemNotFound(item='User')
    if anno is DiscordMember:
        if not ctx.guild:
            raise ItemNotFound(item='Member')
        try:
            if type(item) is inspect.Parameter:
                return item.default
            return await ctx.guild.get_member(re.sub(r'[@!<>]', '', item))
        except HTTPError:
            raise ItemNotFound(item='Member')
    return anno(item)


class CommandsBot(GatewayBot):

    def __init__(self, *args, command_prefix: str, case_sensitivity: bool = False, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.case_sensitivity = case_sensitivity
        self.prefix = command_prefix
        self.__commands: dict[str, ClientCommand] = {}
        self._check: Optional[Callable] = None

        # Default listeners and commands
        self.add_listener(self.message_listener, event='message_create', type='default')
        self.add_command(self.default_help, command='help', description='Shows this message')

    async def message_listener(self, msg: GatewayMessage) -> None:
        """
        Default on_message_create,\n
        register messages as commands\n
        :param msg: :class:`GatewayMessage`
            The message to check for
        """
        if not msg.data.content.startswith(self.prefix) or not (
                msg.data.__dict__.get('guild_id', False) or self.allow_dms) \
                or msg.data.author['id'] == self.user.id:
            return
        await self.register_commands(CommandContext(msg.data, self.api))

    async def default_help(self, ctx: CommandContext) -> None:
        """
        Default help function.\n
        Shows help manu\n
        :param ctx: The given context
        :type ctx: CommandContext
        """
        embed = Embed(title='Bot Commands', color=Color.Blue)
        aliases: list[str] = []

        # Checking aliases
        seen: list[Callable] = []
        for name, command in self.__commands.items():
            if command.function not in seen:
                seen.append(command.function)
            else:
                aliases.append(name)

        commands = {key: self.__commands[key] for key in sorted(self.__commands.keys())}
        for command_name, command in commands.items():
            if command_name in aliases or not command.eval_check(ctx):
                continue
            sig = signature(command.function)
            args = '' if len(list(sig.parameters)) <= 1 else f" ({', '.join(list(sig.parameters.keys())[1:])})"
            embed.add_field(name=command_name.title() + args, value=f'{command.description}', inline=False)
        await ctx.send(embed=embed)

    def raise_error(self, ctx: CommandContext, error: BaseException) -> None:
        """
        Raise an error safely\n
        :param ctx:
            Context for the error
        :param error:
            The error
        """
        if 'command_error' in self._handlers:
            for error_handler in map(lambda handler: handler.function, self._handlers['command_error']):
                threading.Thread(target=asyncio.run, args=(error_handler(ctx, error),)).start()
        else:
            raise error

    def set_check(self, check: Callable) -> None:
        """
        Sets a check\n
        :param check:
            The check to assign
        """
        self._check = check

    def clear_check(self) -> None:
        """
        Resets Checks
        """
        self._check = None

    def add_command(self, func: Callable, command: str, description: str = '', cooldown: int = 0,
                    check: Optional[Callable] = None) -> None:
        """
        Add a command by name.\n
        :param func:
            The function to assign with the command
        :type func: Callable
        :param command:
            The command name.
        :type command: str
        :param description:
            A description for the command
        :type description: str
        :param cooldown:
            The cooldown for the command
        :type cooldown: int
        :param check:
            Local check for the specified command
            will take context as a param
        :type check: Callable
        """
        self.__commands[command] = ClientCommand(func, description or 'No description provided', cooldown)
        if check is not None:
            self.__commands[command].set_check(check)

    def remove_command(self, command: str) -> bool:
        """
        Removes a command by name.\n
        :param command:
            The command to remove
        :type command: str
        :return
            Success or failure
        """
        if command in self.__commands:
            del self.__commands[command]
            return True
        return False

    def command(
            self,
            name: str = '',
            aliases: list = (),
            description: Optional[str] = None,
            cooldown: int = 0,
            check: Optional[Callable[..., bool]] = None
    ) -> Callable[..., ClientCommand]:
        """
        **Decorator**\n
        Used to assign a function with a command\n
        :param name: :class:`str`
            Name for the command
        :param aliases: :class:`list[str]`
            Aliases for the given command
        :param description: :class:`str`
            A description for the command
        :param cooldown: :class:`int`
            The cooldown for the command
        :param check: :class:`Callable`
            Local check for the command, takes context as argument
        :return: :class:`Callable[[{__name__}], None]`
            The nested function
        """

        def decorator(command: Callable) -> ClientCommand:
            """
            **Nested Decorator**\n
            Takes a function and assign with a command\n
            :param command: :class:`Callable`
                The function to assign
            """
            command_name = name or command.__name__
            self.add_command(command, command_name, description, cooldown, check)
            for alias in aliases:
                self.add_command(command, str(alias), description, cooldown, check)
            return self.__commands[command_name]

        return decorator

    async def register_commands(self, msg: CommandContext) -> None:
        """
        Checks for available assigned commands\n
        if true:
        \tExecutes the commands\n
        if false:
        \tIgnores or raises an error depends on your settings\n
        :param msg: :class:`CommandContext`
            The command, data
        """
        command_name: str = msg.content.replace(self.prefix, '', 1).split(' ')[0]

        if not self.case_sensitivity:
            command_name = command_name.lower()

        if command_name in self.__commands:

            command = self.__commands[command_name]
            func = command.function
            allocator = command.error_allocator
            sig = signature(func).parameters

            # Checking process..
            if (self._check and not self._check(msg)) or not command.eval_check(msg):
                e = CheckFailed(user=msg.author.id)
                if allocator:
                    threading.Thread(target=asyncio.run, args=(allocator(msg, e),)).start()
                    return
                return self.raise_error(msg, e)

            non_default = filter(lambda sign: sign.default is sign.empty and sign.kind is not sign.VAR_POSITIONAL,
                                 sig.values())
            args = msg.content.split(' ')[1:len(sig)]
            if len(list(non_default)) - 1 > len(args):
                return self.raise_error(msg, discord_gateway.MissingArgumentsException(command=command_name))
            else:
                args = msg.content.replace(self.prefix, '', 1).split(' ')[1:]
            values = list(sig.values())
            log.info(f'running command {command_name}({", ".join(args)})')
            try:
                args = [
                    (str(arg) if sign.kind != sign.VAR_POSITIONAL else arg)
                    if sign.annotation is sign.empty else await _conv(self, msg, arg, sign.annotation)
                    for arg, sign in zip_longest(args, values[1:], fillvalue=values[-1])
                    if len(values) - 1
                ]
            except Exception as e:
                if allocator:
                    threading.Thread(target=asyncio.run, args=(allocator(msg, e),)).start()
                    return
                return self.raise_error(msg, e)
            command.dispatch(msg, args, bot=self)
        else:
            self.raise_error(msg, CommandNotFound(f'Unknown command {command_name!r}'))



import random
from enum import Enum
from typing import Union, Any, Optional, Iterator

from typing_extensions import Self
from discord_gateway.types import DiscordUser, DiscordRole

__all__ = (
    'Intents',
    'Permission',
    'Color',
    'ChannelType',
    'Status',
    'ActivityType',
    'AllowedMentions'
)


class Intents(Enum):  # Discord intents
    Guilds: int = 1 << 0
    GuildMembers: int = 1 << 1
    GuildBans: int = 1 << 2
    GuildEmojiAndStickers: int = 1 << 3
    GuildIntegrations: int = 1 << 4
    GuildWebhooks: int = 1 << 5
    GuildInvites: int = 1 << 6
    GuildVoiceStates: int = 1 << 7
    GuildPresences: int = 1 << 8
    GuildMessages: int = 1 << 9
    GuildMessageReactions: int = 1 << 10
    GuildStartTyping: int = 1 << 11
    DirectMessages: int = 1 << 12
    DirectMessageReactions: int = 1 << 13
    DirectMessageTyping: int = 1 << 14
    MessageContent: int = 1 << 15
    GuildScheduledEvents: int = 1 << 16
    AutoModerationConfiguration: int = 1 << 20
    AutoModerationExecution: int = 1 << 21
    GuildsAll: int = 69631
    DirectMessagesAll: int = 28672
    AutoModerationAll: int = 3145728
    All: int = 3244031


class Permission(Enum):
    CreateInstantInvites: int = 1 << 0
    KickMembers: int = 1 << 1
    BanMembers: int = 1 << 2
    Administrator: int = 1 << 3
    ManageChannels: int = 1 << 4
    ManageGuild: int = 1 << 5
    AddReactions: int = 1 << 6
    ViewAuditLog: int = 1 << 7
    PrioritySpeaker: int = 1 << 8
    Stream: int = 1 << 9
    ViewChannel: int = 1 << 10
    SendMessages: int = 1 << 11
    SendTTSMessages: int = 1 << 12
    ManageMessages: int = 1 << 13
    EmbedLinks: int = 1 << 14
    AttachFiles: int = 1 << 15
    ReadMessageHistory: int = 1 << 16
    MentionEveryone: int = 1 << 17
    UseExternalEmojis: int = 1 << 18
    ViewGuildInsights: int = 1 << 19
    Connect: int = 1 << 20
    Speak: int = 1 << 21
    MuteMembers: int = 1 << 22
    DeafenMembers: int = 1 << 23
    MoveMembers: int = 1 << 24
    UseVAD: int = 1 << 25
    ChangeNickname: int = 1 << 26
    ManageNicknames: int = 1 << 27
    ManageRoles: int = 1 << 28
    ManageWebhooks: int = 1 << 29
    ManageEmojisAndStickers: int = 1 << 30
    UseApplicationCommands: int = 1 << 31
    RequestToSpeak: int = 1 << 32
    ManageEvents: int = 1 << 33
    ManageThreads: int = 1 << 34
    CreatePublicThreads: int = 1 << 35
    CreatePrivateThreads: int = 1 << 36
    UseExternalStickers: int = 1 << 37
    SendMessagesInThreads: int = 1 << 38
    UseEmbeddedActivities: int = 1 << 39
    ModerateMembers: int = 1 << 40

    @classmethod
    def all(cls) -> int:
        return sum(map(lambda p: p.value, list(cls)))

    @classmethod
    def from_int(cls, number: Union[str, int]) -> Optional[Iterator[Self]]:
        number = int(number)
        if not number:
            return
        for permission in list(cls):
            if number & permission.value:
                yield permission

    def __repr__(self) -> str:
        return self.name


class ChannelType(Enum):
    TextChannel: int = 0
    DirectChannel: int = 1
    VoiceChannel: int = 2
    GroupChannel: int = 3
    CategoryChannel: int = 4
    AnnouncementChannel: int = 5
    AnnouncementThread: int = 10
    PublicThread: int = 11
    PrivateThread: int = 12
    StageChannel: int = 13
    GuildDirectory: int = 14
    ForumChannel: int = 15


class Color(Enum):
    Blue: int = 0x3498DB
    Green: int = 0x2ECC71
    Yellow: int = 0xFEE75C
    Red: int = 0xE74C3C
    Default: int = 0x8b8b8b
    Orange: int = 0xE67E22
    Gray: int = 0x898989
    Gold: int = 0xeb8c26
    Purple: int = 0xb024f1
    Pink: int = 0xff2ebd

    @classmethod
    def random(cls) -> Self:
        return random.choice(list(cls))


class Status(Enum):
    Online: str = 'online',
    Offline: str = 'offline',
    Idle: str = 'idle'
    DND: str = 'dnd'
    Invisible: str = 'invisible'


class ActivityType(Enum):
    Playing: int = 0
    Streaming: int = 1
    Listening: int = 2
    Watching: int = 3
    Custom: int = 4
    Competing: int = 5


class AllowedMentions:
    __slots__ = 'payload'

    def __init__(self, *, everyone: bool = True, roles: Union[list[DiscordRole], bool] = True,
                 users: Union[list[DiscordUser], bool] = True, replied_user: bool = True) -> None:
        self.payload: dict[str, Any] = {'parse': []}

        if everyone:
            self.payload['parse'].append('everyone')
        if roles:
            self.payload['parse'].append('roles')
        if users:
            self.payload['parse'].append('users')

        # Appending to parse

        if type(roles) is list:
            self.payload['roles'] = [role.id for role in roles]

        if type(users) is list:
            self.payload['users'] = [user.id for user in users]

        if replied_user:
            self.payload['replied_user'] = True

    @classmethod
    def all(cls) -> Self:
        return cls(everyone=True, roles=True, users=True, replied_user=True)

    @classmethod
    def none(cls) -> Self:
        return cls(everyone=False, roles=False, users=False, replied_user=False)



from enum import Enum
from types import SimpleNamespace
from typing import Optional

from typing_extensions import Self
from discord_gateway.api import HttpClient

__all__ = (
    'DiscordUser',
    'DiscordRole'
)


class DiscordUser:
    __slots__ = 'id', 'name', 'mention', 'bot'

    def __init__(self) -> None:
        self.id: str = ''
        self.name: str = ''
        self.mention: str = ''
        self.id: bool = True


class DiscordRole:
    __slots__ = 'id', 'name', 'color', 'api', 'data', '__token', 'guild_id', 'permissions', 'position'

    def __init__(self) -> None:
        self.id: str = ''
        self.name: str = ''
        self.permissions: str = ''
        self.color: int = 0
        self.position: int = 0
        self.data: Optional[SimpleNamespace] = None
        self.guild_id: str = ''
        self.api: Optional[HttpClient] = None

    def __repr__(self) -> str:
        return self.name

    async def delete(self) -> None:
        pass

    async def edit(
            self,
            name: Optional[str] = None,
            *,
            permissions: Optional[Enum] = None,
            color: Optional[int] = None,
            hoist: Optional[bool] = None,
            unicode_emoji: Optional[str] = None,
            mentionable: Optional[bool] = None
    ) -> Self:
        pass



"""
Discord Gateway framework

Check github at ...
"""

__title__ = 'DiscordGateway'
__author__ = 'zYoshaa'
__version__ = '2.0.1'

from .api import *
from .constants import *
from .discord_objects import *
from .discord_user import *
from .gatewaybot import *
from .gateway_handler import *
from .gateway import *
from .embed import *
from .request_manager import *
from .presence import *
from.guild_objects import *
from .exceptions import *
import logging as log

log.basicConfig(encoding='utf-8', level=log.ERROR)



import os
from dataclasses import dataclass
from typing import Union, Optional, Any

from discord_gateway import constants

__all__ = (
    'Embed',
    'DiscordFile',
)


@dataclass
class Author:
    __slots__ = 'name', 'url', 'icon_url'

    name: str
    url: Optional[str]
    icon_url: Optional[bytes]


class Embed:
    __slots__ = 'title', 'description', 'url', 'color', 'timestamp', 'fields', 'author', 'footer'

    def __init__(self, title: str, description: str = '', *, url: str = '',
                 color: Union[constants.Color, int] = constants.Color.Default) -> None:
        self.title = title
        self.description = description
        self.url = url
        self.color = color.value
        self.author: Optional[dict] = None
        self.footer: Optional[dict] = None
        self.fields: list[dir] = []

    def add_field(self, name: str, value: str, inline: bool = True) -> None:
        """
        Add a field to the embed\n
        :param name:
            Name of field
        :type name: str
        :param value:
            Value of field
        :type value: str
        :param inline:
            Weather or not the field should be inlined
        :type inline: bool
        """
        self.fields.append(
            {
                'name': name,
                'value': value,
                'inline': inline,
            }
        )

    def set_author(self, name: str, url: Optional[str] = None, icon_url: Optional[str] = None) -> None:
        """
        Sets the author for the embed\n
        :param name:
            Name of author
        :type name: str
        :param url:
            url for author
        :type url: Optional[str]
        :param icon_url:
            Avatar url of author
        :type icon_url: Optional[str]
        """
        author = Author(name, url, bytes(icon_url) if icon_url else None)
        self.author: dict = {
            attr: getattr(author, attr)
            for attr in author.__slots__
            if hasattr(author, attr)
        }

    def set_footer(self, *, content: Optional[str] = None, icon_url: Optional[Any] = None) -> None:
        """
        Sets the footer of the embed\n
        :param content:
            The footer text
        :param icon_url:
            Url for the footer icon
        """
        self.footer = {}

        if content is not None:
            self.footer['text'] = content

        if icon_url:
            self.footer['icon_url'] = str(icon_url)

    def to_json(self) -> dict:
        """
        Converts the embed to dict\n
        :return:
            Dict formatted embed
        """
        return {
            key: getattr(self, key)
            for key in self.__slots__
            if hasattr(self, key)
        }


class DiscordFile:

    def __init__(self, path: Union[str, bytes, os.PathLike], filename: Optional[str] = None) -> None:
        self.path = open(path, 'rb')
        self.filename = filename or self.path.name

    def to_json(self) -> dict:
        return {
            'file': (self.filename, self.path)
        }
